#
# task.rb - Task System
#
require 'singleton'
require 'forwardable'

module PumpkinModern
  class Task
    DefaultGameInfo = Struct.new(:key, :screen)

    DEFAULT_PRIORITY = 5000
    [:add, :clear, :act, :render, :info].each do |m|
      self.instance_eval <<-EOD
        def #{m}(*args)
          Runner.instance.__send__(:#{m}, *args)
        end
      EOD
    end

    def self.inherited(subclass)
      (@@task_classes ||= []).push subclass
    end

    def self.priorities
      ret = {}
      @@task_classes.map{|klass| ret[klass] = klass.const_get(:DEFAULT_PRIORITY)}
      ret
    end

    def self.priority(n)
      const_set(:DEFAULT_PRIORITY, n) #SomeTask::DEFAULT_PRIORITY
    end

    def self.exclusive(flag)
      Runner.instance.exclusives << self if flag
    end

    def self.graphic(key, path, opt = {})
      (@@graphic_list ||= []).push [key, path, self, opt]
    end

    def self.load_graphics
      @@graphic_list.each do |key, path, klass, opt|
        image = SDL::Surface.load_bmp(path)
        if opt
          image = image.copy_rect(opt[:x], opt[:y], opt[:w], opt[:h])
        end
        image.set_color_key(SDL::SRCCOLORKEY, [255, 255, 255])
        klass.class_variable_set("@@#{key}".to_sym, image)
      end
    end

    def act(info)
    end

    def render(info)
    end

    def die
      @is_dead = true
    end

    def dead?
      @is_dead==true
    end

    def alive?
      not dead?
    end

    attr_accessor :is_dead

    def priority
      @priority or self.class.const_get(:DEFAULT_PRIORITY)
    end
    

    class Runner
      include Singleton

      def initialize
        @tasks = []
        @exclusives = []
        @exclusive = nil
        @info = if defined? GameInfo then GameInfo.new else DefaultGameInfo.new end
        Task.load_graphics
      end
      attr_reader :info
      attr_accessor :exclusives

      def num_tasks
        @tasks.size
      end

      def add(*tasks)
        tasks.each{|task| add_task(task) }
      end

      def add_task(task)
        task.is_dead = false
        if @exclusives.include? task.class
          @exclusive = task
        elsif @tasks.find{|t| t.__id__ == task.__id__}
          raise "Task #{task.class} added two times"
        end
        @tasks << task
      end

      def clear
        @tasks.clear
      end

      def act
        if @exclusive
          (@exclusive.dead? ? @exclusive = nil : @exclusive.act(@info))
        else
          @tasks = @tasks.sort_by{|task| task.priority}
          @tasks.reject!{|task| task.dead?}
          @tasks.each{|task| task.act(@info)}
        end
      end

      def render
        @tasks.each{|task| task.render(@info)}
      end

    end # class Runner

  end
end
